// This may look like C code, but it is really -*- C++ -*-
//
//	Verify construction of fractal maps and show them off
//
// $Id: vfractal_map.cc,v 2.2 1996/07/15 19:07:30 oleg Exp oleg $

#include "./image.h"
#include "fractal_map.h"
#include "filter.h"
#include "iostream.h"
#include "MyImageViewer.h"

static void test_nonoise(const card order)
{
  cout << "\nTest constructing a map of order " << order << 
	  " without any noise\n" << endl;

  class ZeroNoise : public FractalMap
  {
  public:
    ZeroNoise(const card order) : FractalMap(order) {}
    ZeroNoise(const card order,const card smoothness,
  	      const card starting_scale_log)
      : FractalMap(order,smoothness,starting_scale_log) {}
    inline int get_noise(const card scale) const { return 0; }
  };

  IMAGE map = ZeroNoise(order);
  const GRAY pixel_mid_val = 1<<(map.q_depth()-1);
  verify_pixel_value(map,pixel_mid_val);
  IMAGE map1 = ZeroNoise(order,0,2);
  verify_pixel_value(map1,pixel_mid_val);
  cout << "\nDone" << endl;
}

static void test_norandom(void)
{
  const card order = 2;
  cout << "\nTest constructing a map of order " << order << 
	  " with non-random noise\n" << endl;

  class NoRNoise : public FractalMap
  {
  public:
    NoRNoise(const card order, const card smoothness,
  	     const card starting_scale_log,
	     const card bits_per_pixel)
      : FractalMap(order,smoothness,starting_scale_log,bits_per_pixel)
      { cout << "\tsmoothness " << smoothness << ", starting scale log"
             << starting_scale_log << ",depth " << bits_per_pixel << endl; }
    inline int get_noise(const card scale) const { return scale/2; }
  };

  {
    IMAGE map = NoRNoise(order,0,1,7);
    verify_pixel_value(map,96);
  }
  {
    IMAGE map = NoRNoise(order,0,2,7);
    verify_pixel_value(map,112);
  }
  {
    IMAGE map = NoRNoise(order,1,2,7);
    verify_pixel_value(map,96);
  }
  
#if 0
  map.print("map");

  IMAGE expected(map);
  expected = 2; expected(0,0) = 0;
  expected.square_of(3,rowcol(1,1)) = 4;
  assert( map == expected );
#endif

  cout << "\nDone" << endl;
}


// Datatypes for the FractalMapWindow demo db

class DBParams;

class PIntItem
{
  const char * const name;
  short value;
public:
  PIntItem(const char _name [], const int default_value)
    	: name(_name), value(default_value) {}
  operator int (void) const		{ return value; }
  const char * q_label(void) const	{ return name; }
  const PIntItem& operator << (const BMessage& db);
  const PIntItem& operator << (const DBParams& db);
};

struct Porder : public PIntItem
{
  Porder(const int value = 8)
  	: PIntItem("order",value) {}
};
struct Psmoothness : public PIntItem
{
  Psmoothness(const int value = 0)
  	: PIntItem("smoothness",value) {}
};
struct Pstarting_scale_log : public PIntItem
{
  Pstarting_scale_log(const int value = 0)
  	: PIntItem("inscale",value) {}
};
  
struct Ppost_process : public PIntItem
{
  enum Code {None = 0, Blur, Smooth_low, Smooth_high};
  Ppost_process(const Code _value = None) : PIntItem("PostProcessing",_value) {}
  void operator () (IMAGE& image) const;
  const Ppost_process& operator << (const DBParams& db)
  { PIntItem::operator<<(db); return *this; }
};
  							// Database of parameters for this window organized
  							// as a menu bar
class DBParams
{
  friend class PIntItem;
  static BMenuBar * mbar;
  void build_initial(void);
  struct RadioMode
  {
    const bool value;
    RadioMode(const bool _val) : value(_val) {}
  };
  struct SetMark
  {
    const int item_index;
    SetMark(void) : item_index(-1) {}
    SetMark(const int _index) : item_index(_index) {}
  };
  class DMenu
  {
    friend class PIntItem;
    BMenu& menu;
    static BMenu& find_submenu(const char name [], const BMenu * mbar);
  public:
    DMenu(const char name []) : menu(*(new BMenu(name))) {}
    DMenu(const char name [], const BMenuBar * mbar) : menu(find_submenu(name,mbar)) {}
    operator BMenu * (void) 	{ return &menu; }
	DMenu& operator << (const PIntItem& item);
	DMenu& operator << (const RadioMode& item) { menu.SetRadioMode(item.value); return *this; }
	DMenu& operator << (const SetMark& item);
  };

public:

  enum { DBPMessage = 100 };

  DBParams(const BMessage& db_ext);		// Internalize the db
  operator BView* (void)	{ return mbar; }
  float Height(void) const	{ return mbar->Bounds().Height(); }
  BMessage * save_state(void) const;
};

class FractalMapWindow : public BWindow
{
  BLooper& parent;
  DBParams paramsdb;
  IMAGE map;

public:
   FractalMapWindow(const BMessage& params, BLooper& _parent);
   
   void	MessageReceived(BMessage *an_event)
   {
     if(an_event->what==DBParams::DBPMessage)
     {
       assert(RemoveChild(paramsdb));
       parent.PostMessage(paramsdb.save_state());	// Parent would relay this message to us
       Quit();										// and resurrect us
     }
   }
   bool QuitRequested(void)
     { assert(RemoveChild(paramsdb));
       parent.PostMessage(B_QUIT_REQUESTED); return true; }
};

FractalMapWindow::FractalMapWindow(const BMessage& params, BLooper& _parent)
   : BWindow(BRect(50,50,80,80),		// just to construct the window; will
   									// be resized later, after map is
   									// constructed
   			  "A Fractal Map",B_TITLED_WINDOW,
   			  B_NOT_ZOOMABLE|B_NOT_RESIZABLE|B_NOT_MINIMIZABLE),
   	parent(_parent),
   	paramsdb(params),
   	map(FractalMap(Porder()<<paramsdb,
   				   Psmoothness()<<paramsdb,
   				   Pstarting_scale_log()<<paramsdb))
   	{
	  Lock();
   	  (Ppost_process()<<paramsdb)(map);
   	  AddChild(paramsdb);
	  const float mb_height = paramsdb.Height();
   	  ResizeTo(map.q_ncols(),map.q_nrows()+mb_height+1);	// Set the proper dimensions
      AddChild(new MyImageView(map,BPoint(0,mb_height)));
  	  Show();
	  Unlock();
   	}


						// will be resized later
BMenuBar * DBParams::mbar = 0;

DBParams::DMenu& DBParams::DMenu::operator << (const PIntItem& item)
{
  BMessage * msg = new BMessage(DBPMessage);
  assert( msg->AddInt16("value",(int)item) == B_OK );
  menu.AddItem(new BMenuItem(item.q_label(),msg));
  return *this;
}

				// Set Check on the given item, or on the last one (if index = -1)
DBParams::DMenu& DBParams::DMenu::operator << (const DBParams::SetMark& item)
{
  BMenuItem * mitem = menu.ItemAt(item.item_index == -1 ?
  						menu.CountItems()-1 : item.item_index);
  assert( mitem != 0 );
  mitem->SetMarked(true);
  return *this;
}

BMenu& DBParams::DMenu::find_submenu(const char name [], const BMenu * mbar)
{
  assert( mbar != 0 );
  const BMenuItem * item = mbar->FindItem(name);
  assert( item != 0 );
  const BMenu * submenu = item->Submenu();
  assert( submenu != 0 );
  return *submenu;
}


						// Build the initial menubar (the database of parameters)
void DBParams::build_initial(void)
{
  mbar = new BMenuBar(BRect(0,0,10,10),"FW Params");
  mbar->AddItem(DMenu(Porder().q_label())
  				<< PIntItem("256x256",8) << SetMark()
  				<< PIntItem("512x512",9)
  				<< PIntItem("1024x1024",10)
  				<< RadioMode(true));
  mbar->AddItem(DMenu(Psmoothness().q_label())
  				<< PIntItem("0",0) << SetMark()
  				<< PIntItem("1",1)
  				<< PIntItem("2",2)
  				<< RadioMode(true));
  mbar->AddItem(DMenu(Pstarting_scale_log().q_label())
  				<< PIntItem("0",0) << SetMark()
  				<< PIntItem("1",1)
  				<< PIntItem("2",2)
  				<< RadioMode(true));
  mbar->AddItem(DMenu(Ppost_process().q_label())
  				<< PIntItem("None",Ppost_process::None) << SetMark()
  				<< PIntItem("Blur",Ppost_process::Blur)
  				<< PIntItem("Smooth a bit",Ppost_process::Smooth_low)
  				<< PIntItem("Smooth more",Ppost_process::Smooth_high)
  				<< RadioMode(true));
}

DBParams::DBParams(const BMessage& db_ext)
{
  if(mbar == 0)
    build_initial();
}

						// Save the current state into a new (globally allocated)
						// message
BMessage * DBParams::save_state(void) const
{
  BMessage * msg = new BMessage(DBPMessage);
  return msg;
}

			// Parameters database access
const PIntItem& PIntItem::operator << (const BMessage& parms)
{
  if( parms.HasInt16(name) )
  {
    assert( parms.FindInt16(name,&value) == B_OK );
  }
  return *this;
}

const PIntItem& PIntItem::operator << (const DBParams& parms)
{
  DBParams::DMenu menu(name,DBParams::mbar);
  if( const BMenuItem * item = ((BMenu *)menu)->FindMarked() )
    value = (PIntItem("value",-1) << *(item->Message()));
  return *this;
}

void Ppost_process::operator () (IMAGE& image) const
{
  switch((int)(*this))
  {
    case None:
    	 return;
    	 
    case Blur:
         filter(image).conv(conv_kernel(1,1,1,CommonDenom(3))).clip_to_intensity_range();
     	 return;
     	 
    case Smooth_low:
     	 filter(image).nondet_average(50);
    	 return;
    	 
    case Smooth_high:
     	 filter(image).nondet_average(100);
    	 return;
    
    default:
    	  assert(0 /*wrong code*/);
  }
}

#if 0
				// Show off maps
static void demo_maps(void)
{
  class GaussNoise : public FractalMap
  {
  public:
    GaussNoise(const card order, const card smoothness,
  	     const card starting_scale_log)
      : FractalMap(order,smoothness,starting_scale_log) {}
				// Well-known Gaussian random number generator
				// Note rand() returns a number [0,2^15-1],
				// that is, within [0,1] with a 15-bit implicit
				// fraction
    inline int get_noise(const card scale) const {
      long sum = 0;
      for(register int i=0; i<12; i++)
	sum += rand();			// keep the result within 
      return (scale * (sum-(6<<15)))>>17; } // [-scale/2,scale/2]
  };

  WindowDestroyer window_destroyer;
  
  for(;;)
  {
    cout << "Enter a map's order (log2 of map's dimension) (0 to quit)" 
      << endl;
    int order;
    cin >> order;
    if( order <= 1 )
     return;
    cout << "Select random number generator: 0 - Uniform, 1 - Gaussian"
         << endl;
    int type;
    cin >> type;
    cout << "Enter smoothness and the starting scale log (both can be zero)"
         << endl;
    int smoothness, starting_scale_log;
    cin >> smoothness >> starting_scale_log;
    IMAGE map = type == 0 ? 
      (LazyImage&)FractalMap(order,smoothness,starting_scale_log) :
      (LazyImage&)FractalMap(order,smoothness,starting_scale_log);
    cout << "Select post-processing option" << endl;
    cout << "\t0 - no postprocessing" << endl;
    cout << "\t1 - apply blurring" << endl;
    cout << "\t2 - smooth non-deterministically" << endl;
    cin >> type;
    if( type == 1 )
      filter(map).conv(conv_kernel(1,1,1,CommonDenom(3))).
	clip_to_intensity_range();
    else if( type == 2 )
      filter(map).nondet_average(50);

    char buffer[70];
    sprintf(buffer,"smoothness %d, starting_scale_log %d",
    	    smoothness,starting_scale_log);
    window_destroyer.add(new MyImageWindow(map,buffer));
//    MyImageViewer image_viewer(map,buffer);
//    map.write_pgm("/tmp/abc",buffer);
//    system("/usr/local/bin/X11/xv /tmp/abc");
//    map.display(buffer);
  }
}

#endif

int main(void)
{
  test_nonoise(4); 
  test_nonoise(9); 
  test_norandom();

  srand(time(0));
  rand(); rand(); rand(); rand();

class MyImageViewer : public BApplication
{
public:
	    MyImageViewer(void)
	    : BApplication("application/x-vnd.x-vfractal-map")
	    { }
   		void MessageReceived(BMessage *an_event)
   		{
     	  if(an_event->what==DBParams::DBPMessage)
	        new FractalMapWindow(*an_event,*this);
        }
	    void ReadyToRun(void)
	    {
	      new FractalMapWindow(
	      		BMessage(DBParams::DBPMessage),*this);
	    }
	    
};
MyImageViewer().Run();

printf("\nDone\n\n");
return 0;
 
}
